/*
Copyright (c) 2010, Geomatics and Cartographic Research Centre, Carleton 
University
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 - Neither the name of the Geomatics and Cartographic Research Centre, 
   Carleton University nor the names of its contributors may be used to 
   endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

$Id$
*/
package ca.carleton.gcrc.progress;

import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

/**
 * Singleton class that tracks all progress information
 * within a process. It tracks activities based on identifiers. Activities
 * have a description, a total count and a current count. 
 * <br/>
 * The complexity of this class stems from the fact that memory should not be 
 * leaked. Therefore, two settings are used to clean the cache:<ol>
 * <li>Elapsed time since last request for an activity (lastRequestFlushInMs)</li>
 * <li>Maximum number of tracked activities (maxConcurrentActivities)</li>
 * </ol>
 *
 */
public class ProgressTracker {
	
	/**
	 * Default number of milliseconds that an activity is
	 * tracked after the last request for its value.
	 */
	static final long LAST_REQUEST_FLUSH_IN_MS = 5 * 60 * 1000; // 5 minutes
	
	/**
	 * Maximum number of concurrent activities tracked at once.
	 */
	static final long MAX_CONCURRENT_ACTIVITIES = 1000; // at most a 1000 activities

	static private ProgressTracker g_singleton = new ProgressTracker();
	static public ProgressTracker getProgressTracker() {
		return g_singleton;
	}

	/**
	 * Creates and return a unique identifier that can be used to track
	 * activities.
	 * @return A unique identifier.
	 */
	static public String createIdentifier() {
		ProgressTracker tracker = getProgressTracker();
		
		return tracker.getUniqueIdentifier();
	}
	
	/**
	 * This is used to report a new activity on which progress is tracked.
	 * The identifier must be unique across all usage. To ensure that the
	 * identifier is unique, {@link #createIdentifier()} can be used, but it
	 * is not necessary.
	 * @param identifier Unique identifier used to track an activity
	 * @param description Description of the tracked activity
	 * @param totalCount Total count in the activity
	 */
	static public void initProgess(String identifier, String description, long totalCount) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);
		
		synchronized(info) {
			info.setDescription(description);
			info.setTotalCount(totalCount);
		}
	}

	/**
	 * This is used to update the progress of an activity. The identifier used
	 * must match one that was already initialized in {@link #initProgess(String, String, long)}.
	 * @param identifier Unique identifier used to track an activity
	 * @param currentCount Count up to date of this activity
	 */
	static public void updateProgess(String identifier, long currentCount) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);
		
		synchronized(info) {
			info.setCurrentCount(currentCount);
		}
	}

	/**
	 * This is used to set data associated with a an activity. The identifier used
	 * must match one that was already initialized in {@link #initProgess(String, String, long)}.
	 * @param identifier Unique identifier used to track an activity
	 * @param currentCount Count up to date of this activity
	 */
	static public void updateProgessData(String identifier, Map<String,String> data) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);
		
		synchronized(info) {
			info.setData(data);
		}
	}
	
	/**
	 * Before an activity is complete, it is possible to start a new activity and chain
	 * it the a previous one. This allows tracking objects to follow the path of an execution.
	 * The identifier used
	 * must match one that was already initialized in {@link #initProgess(String, String, long)}.
	 * @param identifier Unique identifier used to track an activity
	 * @param chainedActivity Activity that is being chained (follow) this one
	 */
	static public void addProgessChain(String identifier, ProgressInfo chainedActivity) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);
		
		synchronized(info) {
			info.addChainedActivity(chainedActivity);
		}
	}
	
	/**
	 * This is used to report a that an activity has completed.
	 * The identifier must be unique across all usage. To ensure that the
	 * identifier is unique, {@link #createIdentifier()} can be used, but it
	 * is not necessary.
	 * @param identifier Unique identifier used to track an activity
	 * @param errorMessage Error message if an error occurred. False otherwise.
	 */
	static public void completeProgess(String identifier, String errorMessage) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);
		
		synchronized(info) {
			info.setCompleted(true);
			info.setErrorMessage(errorMessage);
			
			// If no error, bump up count to 100%
			if( null == errorMessage ) {
				info.setCurrentCount( info.getTotalCount() );
			}
		}
	}

	/**
	 * Returns the current status of an activity. This call
	 * is expensive, since it performs clean up. Also, when accessing
	 * the progress info (returned object), it should be synchronized
	 * since it is shared with other threads.
	 * @param identifier Unique identifier used to track an activity
	 * @return Object that contains information about a tracked activity
	 */
	static public ProgressInfo getProgess(String identifier) {
		ProgressTracker tracker = getProgressTracker();
		return tracker.getProgressInfoUpdateLastRequest(identifier);
	}
	
	private long count = 0;
	private Map<String,ProgressInfo> infoMap = new HashMap<String,ProgressInfo>();
	private List<ProgressInfo> infoOrderedLastAccess = new Vector<ProgressInfo>();
	private long lastRequestFlushInMs;
	private long maxConcurrentActivities;
	
	// Singleton
	private ProgressTracker() {
		lastRequestFlushInMs = LAST_REQUEST_FLUSH_IN_MS;
		maxConcurrentActivities = MAX_CONCURRENT_ACTIVITIES;
	}

	public synchronized long getLastRequestFlushInMs() {
		return lastRequestFlushInMs;
	}

	public synchronized void setLastRequestFlushInMs(long lastRequestFlushInMs) {
		this.lastRequestFlushInMs = lastRequestFlushInMs;
	}

	public synchronized long getMaxConcurrentActivities() {
		return maxConcurrentActivities;
	}

	public synchronized void setMaxConcurrentActivities(long maxConcurrentActivities) {
		this.maxConcurrentActivities = maxConcurrentActivities;
	}
	
	synchronized private String getUniqueIdentifier() {
		String result = "progress"+count;
		++count;
		return result;
	}

	synchronized private ProgressInfo getProgressInfoFromId(String identifier) {
		ProgressInfo info = infoMap.get(identifier);
		if( null == info ) {
			info = new ProgressInfo(identifier);
			infoMap.put(identifier, info);
			infoOrderedLastAccess.add(info);
			
			// Do not cross max
			if( maxConcurrentActivities > 0 ) {
				while( infoOrderedLastAccess.size() > maxConcurrentActivities ) {
					ProgressInfo infoToDelete = infoOrderedLastAccess.get(0);
					deleteInfo(infoToDelete);
				}
			}
		}
		return info;
	}

	private ProgressInfo getProgressInfoUpdateLastRequest(String identifier) {
		ProgressTracker tracker = getProgressTracker();
		ProgressInfo info = tracker.getProgressInfoFromId(identifier);

		Date current = new Date();
		
		synchronized(info) {
			info.setLastRequested(current);
		}

		// Move to back of list
		synchronized(this) {
			infoOrderedLastAccess.remove(info);
			infoOrderedLastAccess.add(info);

			cleanInfo(current);
		}
		
		return info;
	}
	
	synchronized private void deleteInfo(ProgressInfo info) {
		infoMap.remove(info.getIdentifier());
		infoOrderedLastAccess.remove(info);
	}

	synchronized private void cleanInfo(Date currentDate) {
		if( lastRequestFlushInMs <= 0 ) {
			return;
		}
		
		long currentTimeInMs = currentDate.getTime();
		
		while( infoOrderedLastAccess.size() > 0 ) {
			ProgressInfo info = infoOrderedLastAccess.get(0);
			long lastRequestTimeInMs = info.getLastRequested().getTime();
			if( (currentTimeInMs - lastRequestTimeInMs) > lastRequestFlushInMs ) {
				deleteInfo(info);
			} else {
				return;
			}
		}
	}

}
